/* eslint-disable react-native/no-inline-styles */
import {FlatList, PanResponder, StyleSheet, View} from 'react-native';
import React, {createRef, RefObject, useEffect, useRef, useState} from 'react';
import Circle from './Circle';
import Scale, {ScaleAnimationHandle} from './transitions/Scale';
import Fade, {FadeAnimationHandle} from './transitions/Fade';
import {CopilotCameraSlotHook} from '@s20.ai/copilot-camera-sdk';
import {Slot} from '@s20.ai/copilot-camera-sdk/src/types/AOP';
import {Button} from 'react-native-paper';

/**
 *
 * @param {number} slotHeight
 * @param {number} margin
 * @returns {number} height
 * @description
 * used to calculate the height
 * of the container containing
 * the slots
 * (also accounts for space between the slot items)
 */
const slotsContainerHeightCalculator = (
  slotHeight: number,
  margin: number,
): number => {
  return 9 * slotHeight + 7 * margin;
};

/**
 *
 * @param {number} slotHeight
 * @param {number} margin
 * @returns {number} offset
 * @description
 * used to calculate the start
 * and end offset to position
 * active element at the center
 * of the container
 */
const offsetToPositionCenter = (slotHeight: number, margin: number) => {
  return slotHeight * 4 + 5 * margin;
};

/**
 *
 * @param {number} activeIndex
 * @param {number} currentIndex
 * @returns {number} scale
 * @description
 * used to calculate scale value of list item based on a
 * comparison between active index in the list
 * and current index of given item */
const scaleFactor = (activeIndex: number, currentIndex: number) => {
  if (activeIndex - currentIndex === 0) {
    return 1;
  } else if (Math.abs(activeIndex - currentIndex) < 5) {
    return 1 - 0.1 * Math.abs(activeIndex - currentIndex);
  } else {
    return 0.5;
  }
};

/**
 *
 * @param {number} activeIndex
 * @param {number} currentIndex
 * @returns {number} opacity
 * @description
 * used to calculate opacity value of list item based on a
 * comparison between active index in the list
 * and current index of given item
 */
const opacityFactor = (activeIndex: number, currentIndex: number): number => {
  if (activeIndex - currentIndex === 0) {
    return 1;
  } else if (Math.abs(activeIndex - currentIndex) < 5) {
    return 1 - 0.2 * Math.abs(activeIndex - currentIndex);
  } else {
    return 0.2;
  }
};

/**
 * @interface
 * @param {number} slotItemHeight
 * @param {number} slotItemSpacing
 */
export interface SafecamSlotList {
  slotItemHeight?: number;
  slotItemSpacing?: number;
}

/**
 * @component
 * Represents a slotList component instance
 * @param {number} slotItemHeight
 * @param {number} slotItemSpacing
 * @description
 * displays the list of available slots
 * and lets the user change the slot
 * using a "slide" gesture
 * @example
 * <SlotList slotItemHeight={30} slotItemSpacing={5}/>
 * @returns {React.FC} SlotList
 */
const SlotList: React.FC<CopilotCameraSlotHook> = (
  props: CopilotCameraSlotHook,
) => {
  const slotItemHeight = 30;
  const slotItemSpacing = 4;

  const {slots, activeSlot, activeSlotIndex, setActiveSlot} = props;

  // to record and set the list container height
  // used for calculating initial and final item position offsets
  const [containerHeight, setContainerHeight] = useState<number>(0);

  // listView ref used for triggering declarative scroll actions
  const ref = useRef<FlatList>(null);

  // contains the actual gesture value ("right" | "left")
  // with a timestamp applied to allow for re-setting of same
  // type of gesture
  const [gesture, setGesture] = useState<string>('');

  // contains width value for each button, (since due to different word length)
  // center position via offset requires unique calculation
  // for each button
  const [buttonWidths, setButtonWidths] = useState<number[]>([]);

  // swipe sensitivity used, this is used to trigger swipe
  // only after the finger has moved across a certain threshold of distance
  const sensitivity = 30;

  // contains animation handle refs for each animated Component wrapper on list item
  // to allow for transitions to be triggered declaratively
  const [scaleRefs, setScaleRefs] = useState<RefObject<ScaleAnimationHandle>[]>(
    [],
  );
  const [fadeRefs, setFadeRefs] = useState<RefObject<FadeAnimationHandle>[]>(
    [],
  );

  const panResponder = useRef(
    PanResponder.create({
      onMoveShouldSetPanResponder: () => true,
      onMoveShouldSetPanResponderCapture: () => true,
      onPanResponderRelease: (_, gestureState) => {
        // used to calculate if the user has swiped up/down
        if (gestureState.dy > sensitivity) {
          setGesture(`down%SEPARATE%${Date.now()}`);
        } else if (gestureState.dy < -sensitivity) {
          setGesture(`up%SEPARATE%${Date.now()}`);
        }
      },
      onPanResponderTerminationRequest: () => true,
    }),
  ).current;

  // used to set the ref handle placeholders
  // for animation on each item to be triggered on component mount
  useEffect(() => {
    setScaleRefs([
      ...Array(slots.length)
        .fill(null)
        .map((_, __) => createRef<ScaleAnimationHandle>()),
    ]);
    setFadeRefs([
      ...Array(slots.length)
        .fill(null)
        .map((_, __) => createRef<FadeAnimationHandle>()),
    ]);
    setButtonWidths([
      ...Array(slots.length)
        .fill(null)
        .map((_, __) => 0),
    ]);
  }, [slots]);

  // used to run a effect, based on the gesture user has selected
  useEffect(() => {
    const g = gesture.split('%SEPARATE%')[0];
    // edge case to allow up swipe to move to next slot only if its not the last slot in state list
    // edge case to allow down swipe to move to first slot only if its not the first slot in state list
    if (g === 'up') {
      setActiveSlot(activeSlotIndex + 1);
    } else if (g === 'down') {
      setActiveSlot(activeSlotIndex - 1);
    }
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [gesture]);

  // effect to display on the UI which slot is active and in center
  useEffect(() => {
    setImmediate(() => {
      if (ref.current) {
        if (containerHeight > 0 && slots.length > 0 && activeSlot) {
          // makes the listView declaratively scroll onto the given index
          // with a calculated offset
          ref.current.scrollToIndex({
            index: activeSlotIndex,
            viewOffset: offsetToPositionCenter(slotItemHeight, slotItemSpacing),
          });
        } else if (activeSlotIndex === 0) {
          ref.current.scrollToIndex({
            index: activeSlotIndex,
            viewOffset: offsetToPositionCenter(slotItemHeight, slotItemSpacing),
          });
        }
        // triggers animation on each slot declaratively based on its index
        scaleRefs.forEach((scaleRef, i) => {
          if (scaleRef.current) {
            scaleRef.current.triggerAnimation(
              scaleFactor(activeSlotIndex as number, i),
            );
          }
        });
        fadeRefs.forEach((fadeRef, i) => {
          if (fadeRef.current) {
            fadeRef.current.triggerAnimation(
              opacityFactor(activeSlotIndex as number, i),
            );
          }
        });
      }
    });

    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, [activeSlot]);

  const defineSlotListItemColor = (slot: Slot) => {
    if (slot.isUploaded) {
      return 'rgba(38, 49, 147, 0.6)';
    } else if (slot.isComplete) {
      return 'rgba(38, 147, 49, 0.6)';
    } else {
      return 'white';
    }
  };

  return (
    <View
      {...panResponder.panHandlers}
      // used to record width of container in which this slot list resides
      onLayout={e => setContainerHeight(e.nativeEvent.layout.height)}
      style={{
        ...styles.root,
        height: slotsContainerHeightCalculator(slotItemHeight, slotItemSpacing),
      }}>
      <FlatList
        ref={ref}
        showsHorizontalScrollIndicator={false}
        showsVerticalScrollIndicator={false}
        data={slots}
        keyExtractor={item => item.slotId}
        scrollEnabled={false}
        onScrollToIndexFailed={() => console.log('scroll error')}
        nestedScrollEnabled={false}
        contentContainerStyle={styles.list}
        onLayout={() => {
          setImmediate(() => {
            ref.current?.scrollToIndex({
              index: 0,
              viewOffset: offsetToPositionCenter(
                slotItemHeight,
                slotItemSpacing,
              ),
            });
          });
        }}
        renderItem={data => {
          return (
            <View
              key={data.item.title}
              style={{
                ...styles.buttonContent,
                height: 50,
                // sets a initial offset from top or bottom based on position of element
                marginBottom:
                  data.index === slots.length - 1
                    ? offsetToPositionCenter(slotItemHeight, slotItemSpacing)
                    : slotItemSpacing,
                marginTop:
                  data.index === 0
                    ? offsetToPositionCenter(slotItemHeight, slotItemSpacing)
                    : 0,
              }}>
              <Scale
                ref={scaleRefs[data.index]}
                config={{
                  duration: 300,
                  startScale: 0,
                  endScale: scaleFactor(activeSlotIndex as number, data.index),
                }}>
                <Fade
                  ref={fadeRefs[data.index]}
                  config={{
                    duration: 300,
                    startOpacity: 0,
                    endOpacity: opacityFactor(
                      activeSlotIndex as number,
                      data.index,
                    ),
                  }}>
                  <Circle
                    style={{
                      marginRight: 3,
                    }}
                    radius={slotItemHeight - 10}
                    // shows the element color based on its progress and active position
                    backgroundColor={defineSlotListItemColor(data.item as Slot)}
                  />
                </Fade>
              </Scale>
              <Button
                onLayout={e => {
                  const widthState = buttonWidths;
                  widthState[data.index] = e.nativeEvent.layout.width;
                  setButtonWidths([...widthState]);
                }}
                style={{
                  opacity: data.item.slotId === activeSlot?.slotId ? 1 : 0,
                  backgroundColor: 'rgba(0, 0, 0, 0.5)',
                }}
                textColor={'white'}>
                {data.item.title}
              </Button>
            </View>
          );
        }}
      />
    </View>
  );
};

export default React.memo(SlotList);

const styles = StyleSheet.create({
  root: {
    zIndex: 3,
    position: 'absolute',
    left: 5,
    alignSelf: 'center',
    top: '30%',
  },
  buttonContent: {
    flexDirection: 'row',
    alignItems: 'center',
    justifyContent: 'space-between',
  },
  list: {
    alignItems: 'flex-start',
    justifyContent: 'center',
  },
});
